           (*-------------------------------------------*
            |  The Dining Mathematicians in CSP-Prover  |
            |               August 2004                 |
            |             December 2004 (modified)      |
            |             November 2005 (modified)      |
            |        Yoshinao Isobe (AIST JAPAN)        |
            *-------------------------------------------*)

theory DM4_Spc_def = DM3_hide:

(*****************************************************************

         1. defines Spc 
         2.
         3. 
         4. 

 *****************************************************************)

(*********************************************************
                     specification 
 *********************************************************)

datatype SpcName = TH0_TH1 | EAT0_TH1 | TH0_EAT1

consts
  SpcDef :: "(SpcName, Event) procDef"

recdef SpcDef "{}"
   "SpcDef
      (X, TH0_TH1)  = ! x:OBS -> IF (x = Eat0) THEN (X EAT0_TH1)
                                 ELSE IF (x = Eat1) THEN (X TH0_EAT1)
                                                    ELSE (X TH0_TH1)"

   "SpcDef
      (X, EAT0_TH1) = ! x:(OBS - {Eat1}) -> IF (x = End0) 
                                            THEN (X TH0_TH1)
                                            ELSE (X EAT0_TH1)"

   "SpcDef
      (X, TH0_EAT1) = ! x:(OBS - {Eat0}) -> IF (x = End1) 
                                            THEN (X TH0_TH1)
                                            ELSE (X TH0_EAT1)"

syntax
  "_SpcFun" :: "(SpcName, Event) procFun" ("SpcFun")
  "_SpcFix" :: "SpcName => Event proc"    ("SpcFix")

translations
    "SpcFun" == "curry SpcDef"
  "SpcFix n" == "(FIX! SpcFun) n"

consts 
  Spc :: "Event proc"

defs
  Spc_def: "Spc == (SpcFix TH0_TH1)"

(*********************************************************
        Lemmas for   ALL n.  verify Spc <=sf Imp n
 *********************************************************)

(*** relation between SpcDef and ImpDef ***)

consts 
  Spc_to_Imp :: "SpcName => Event proc"

primrec
  "Spc_to_Imp
     TH0_TH1 = 
       !<NUM> n .. ((ImpFix TH0) |[CH0]| (ImpFix (VAR n))
                                 |[CH1]| (ImpFix TH1)) -- (CH0 Un CH1)
       |~|
       !<NUM> n .. (Back0 -> (ImpFix TH0) |[CH0]| (ImpFix (VAR n))
                                          |[CH1]| (ImpFix TH1)) -- (CH0 Un CH1)
       |~|
       !<NUM> n .. ((ImpFix TH0) |[CH0]| (ImpFix (VAR n)) 
                                 |[CH1]| Back1 -> (ImpFix TH1)) -- (CH0 Un CH1)
       |~|
       !<NUM> n .. (Back0 -> (ImpFix TH0) |[CH0]| (ImpFix (VAR n))
                                          |[CH1]| Back1 -> (ImpFix TH1)) 
                  -- (CH0 Un CH1)
       |~|
       !<NUM> n:EVENs .. (Eat0 -> (ImpFix (EAT0 n)) |[CH0]| (ImpFix (VAR n))
                                           |[CH1]| Back1 -> (ImpFix TH1))
                        -- (CH0 Un CH1)
       |~|
       !<NUM> n:ODDs  .. (Back0 -> (ImpFix TH0) |[CH0]| (ImpFix (VAR n))
                                                |[CH1]| Eat1 -> (ImpFix (EAT1 n)))
                        -- (CH0 Un CH1)
       |~|
       !<NUM> n:EVENs .. (Eat0 -> (ImpFix (EAT0 n)) |[CH0]| (ImpFix (VAR n))
                                           |[CH1]| (ImpFix TH1))
                        -- (CH0 Un CH1)
       |~|
       !<NUM> n:ODDs  .. ((ImpFix TH0) |[CH0]| (ImpFix (VAR n))
                                       |[CH1]| Eat1 -> (ImpFix (EAT1 n)))
                        -- (CH0 Un CH1)
       |~|
       !<NUM> n:EVENs .. (WR0 (n div 2) -> (ImpFix TH0) |[CH0]| (ImpFix (VAR n))
                                                        |[CH1]| (ImpFix TH1))
                       -- (CH0 Un CH1)
       |~|
       !<NUM> n:EVENs .. (WR0 (n div 2) -> (ImpFix TH0) |[CH0]| (ImpFix (VAR n))
                                                        |[CH1]| Back1 -> (ImpFix TH1))
                       -- (CH0 Un CH1)
       |~|
       !<NUM> n:ODDs  .. ((ImpFix TH0) |[CH0]| (ImpFix (VAR n))
                                       |[CH1]| WR1 ! (3 * n + 1) -> (ImpFix TH1))
                       -- (CH0 Un CH1)
       |~|
       !<NUM> n:ODDs  .. (Back0 -> (ImpFix TH0) |[CH0]| (ImpFix (VAR n))
                                                |[CH1]| WR1 ! (3 * n + 1) -> (ImpFix TH1))
                       -- (CH0 Un CH1)"

  "Spc_to_Imp
     EAT0_TH1 = 
       !<NUM> n:EVENs .. ((ImpFix (EAT0 n)) |[CH0]| (ImpFix (VAR n)) 
                                            |[CH1]| (ImpFix TH1)) 
                       -- (CH0 Un CH1)
       |~|
       !<NUM> n:EVENs .. ((ImpFix (EAT0 n)) |[CH0]| (ImpFix (VAR n))
                                            |[CH1]| Back1 -> (ImpFix TH1))
                       -- (CH0 Un CH1)"

  "Spc_to_Imp
     TH0_EAT1 = 
       !<NUM> n:ODDs .. ((ImpFix TH0) |[CH0]| (ImpFix (VAR n)) |[CH1]| (ImpFix (EAT1 n)))
                       -- (CH0 Un CH1)
       |~|
       !<NUM> n:ODDs .. (Back0 -> (ImpFix TH0) |[CH0]| (ImpFix (VAR n))
                                               |[CH1]| (ImpFix (EAT1 n)))
                       -- (CH0 Un CH1)"

(*********************************************************
                gProc lemmas (routine work)
 *********************************************************)

lemma Spc_gProcFun[simp]:
      "SpcFun : gProcFun"
by (simp add: CSP_Fun_def, rule allI, induct_tac p, simp_all)+

end
