           (*-------------------------------------------*
            |        CSP-Prover on Isabelle2005         |
            |               February 2006               |
            |                                           |
            |        Yoshinao Isobe (AIST JAPAN)        |
            *-------------------------------------------*)

theory DIV_Example = FNF_F:

(*  The following simplification rules are deleted in this theory file *)
(*  because they unexpectly rewrite UnionT and InterT.                 *)
(*                  disj_not1: (~ P | Q) = (P --> Q)                   *)

declare disj_not1 [simp del]

(*  The following simplification rules are deleted in this theory file *)
(*       P (if Q then x else y) = ((Q --> P x) & (~ Q --> P y))        *)

declare split_if  [split del]

(*****************************************************************

         1.
         2. 
         3. 

 *****************************************************************)


consts 
  Count :: "(nat, nat) procDef"

recdef Count "{}"
  "Count  (X, n) = (n -> (X (Suc n))) -- {n}"

declare Count.simps [simp del]

consts
  CountFun :: "(nat, nat) procFun" ("CountFun")
  CountFix :: "nat => nat proc"    ("CountFix")

defs
  CountFun_def: "CountFun == curry Count"
  CountFix_def: "CountFix n == (FIX CountFun) n"

lemma ALL_Count_DIV:
  "ALL m. (CountFun^n) (%y. DIV) (m) =F DIV"
apply (induct_tac n)
apply (simp)
apply (simp)
apply (rule allI)
apply (rule cspF_rw_left)
apply (subgoal_tac 
  "CountFun ((CountFun ^ n) (%y. DIV)) m =F CountFun (%y. DIV) m")
apply (simp)
 apply (simp add: CountFun_def)
 apply (simp add: Count.simps)
 apply (rule cspF_decompo)
 apply (simp)
 apply (rule cspF_decompo)
 apply (simp)
 apply (simp add: curry_def)
 apply (simp add: Count.simps)

apply (simp add: CountFun_def)
apply (simp add: Count.simps)

apply (rule cspF_rw_left)
apply (rule cspF_decompo)
apply (simp)
apply (rule cspF_step)

apply (rule cspF_rw_left)
apply (rule cspF_step)
apply (simp)

apply (rule cspF_rw_left)
apply (rule cspF_IF_False)

apply (rule cspF_rw_left)
apply (rule cspF_decompo)
apply (rule cspF_decompo)
apply (rule cspF_step[THEN cspF_sym])
apply (rule cspF_reflex)
apply (rule cspF_Rep_int_choice_singleton)

apply (rule cspF_rw_left)
apply (rule cspF_decompo)
apply (rule cspF_idem)
apply (rule cspF_reflex)

apply (rule cspF_rw_left)
apply (rule cspF_unit)
apply (rule cspF_DIV_Hiding_Id)
done

lemma Count_DIV:
  "(CountFun^n) (%y. DIV) (m) =F DIV"
apply (simp add: ALL_Count_DIV)
done

lemma CountFIX_DIV:
  "(CountFix) 0 =F !nat n .. DIV"
apply (simp add: CountFix_def)
apply (simp add: FIX_def)
apply (simp add: FIXn_def)
apply (rule cspF_decompo)
apply (simp)
apply (simp add: Count_DIV)
done

(*** full normalising ***)

lemma CountFIX_DIV_Xnorm:
  "(CountFix) 0 =F !nat n .. NDIV"
apply (rule cspF_rw_left)
apply (rule CountFIX_DIV)
apply (rule cspF_decompo)
apply (simp)
apply (rule cspF_NDIV_eqF)
done

(*** in extended full normal form ***)

lemma DIV_Xnorm_in:
  "!nat n .. NDIV : XfnfF_proc"
apply (simp add: XfnfF_proc_def)
apply (rule_tac x="(%n. NDIV)" in exI)
apply (simp)
apply (rule allI)
apply (rule cspF_rw_right)
apply (rule cspF_Dist)
apply (simp)

apply (rule cspF_rw_right)
apply (rule cspF_Rep_int_choice_unit_fun)
apply (simp)

apply (rule cspF_rw_right)
apply (rule cspF_decompo)
apply (simp)
apply (rule cspF_NDIV_eqF[THEN cspF_sym])

apply (rule cspF_rw_right)
apply (rule cspF_DIV_Depth_rest)
apply (rule cspF_NDIV_eqF[THEN cspF_sym])
done

(*** unwinding test ***)

lemma CountFIX_DIV_unwinding_test:
  "(CountFix) n =F (n -> (CountFix) (Suc n)) -- {n}"
apply (simp add: CountFix_def)
apply (rule cspF_rw_left)
apply (rule cspF_unwind_cpo)

 (* ProcFun *)
 apply (simp add: CountFun_def)
 apply (simp add: CSP_Fun_def)
 apply (rule allI)
 apply (induct_tac p)
 apply (simp add: Count.simps)
 apply (simp add: Count.simps)

apply (simp add: CountFun_def)
apply (simp add: Count.simps)
done

(****************** to add them again ******************)

declare split_if    [split]
declare disj_not1   [simp]

end
